﻿namespace FaceDetection
{
  using System;
  using System.Collections.Generic;
  using System.IO;
  using System.Xml;
  using FaceDetection.Convolution;
  using SharpNeat.Core;
  using SharpNeat.Decoders;
  using SharpNeat.Decoders.HyperNeat;
  using SharpNeat.Genomes.HyperNeat;
  using SharpNeat.Genomes.Neat;
  using SharpNeat.Network;
  using SharpNeat.Phenomes;

  static class Common
  {
    public static readonly int WindowSize = 20;

    public static IBlackBox LoadNetwork(string path)
    {
      Console.WriteLine("Loading network {0}", Path.GetFileNameWithoutExtension(path));
      var i = Convert.ToInt32(Path.GetFileNameWithoutExtension(path).Split('_')[0]);
      if (i < 8)
      {
        NeatGenome genome;
        using (XmlReader reader = XmlReader.Create(path))
        {
          reader.ReadToDescendant("Network");
          genome = NeatGenomeXmlIO.ReadGenome(reader, true);
        }
        var decoder = (ConvGenomeDecoder)GetDecoders()[i]();
        return decoder.Decode(genome, GetFactories()[i]().ActivationFnLibrary);
      }
      else
      {
        NeatGenome genome;
        using (XmlReader reader = XmlReader.Create(path))
        {
          genome = NeatGenomeXmlIO.ReadCompleteGenomeList(reader, true, GetFactories()[i]())[0];
        }
        return GetDecoders()[i]().Decode(genome);
      }
    }

    public static Func<NeatGenomeFactory>[] GetFactories()
    {
      return new Func<NeatGenomeFactory>[]
      {
        () => CreateNeatFactory(25, 16),
        () => CreateNeatFactory(16, 25),
        () => CreateNeatFactory(36, 25),
        () => CreateNeatFactory(4, 100),
        () => CreateRbfNeatFactory(25, 16),
        () => CreateRbfNeatFactory(16, 25),
        () => CreateRbfNeatFactory(36, 25),
        () => CreateRbfNeatFactory(4, 100),
        () => CreateHyperNeatFactory(),
        () => CreateHyperNeatFactory(),
        () => CreateHyperNeatFactory(),
        () => CreateHyperNeatFactory(),
      };
    }

    public static Func<IGenomeDecoder<NeatGenome, IBlackBox>>[] GetDecoders()
    {
      return new Func<IGenomeDecoder<NeatGenome, IBlackBox>>[]
      {
        () => new ConvGenomeDecoder(WindowSize, 4, 4),
        () => new ConvGenomeDecoder(WindowSize, 5, 5),
        () => new ConvGenomeDecoder(WindowSize, 5, 3),
        () => new ConvGenomeDecoder(WindowSize, 10, 10),
        () => new ConvGenomeDecoder(WindowSize, 4, 4),
        () => new ConvGenomeDecoder(WindowSize, 5, 5),
        () => new ConvGenomeDecoder(WindowSize, 5, 3),
        () => new ConvGenomeDecoder(WindowSize, 10, 10),
        () => CreateHyperNeatDecoder(1, 4, 4, true),
        () => CreateHyperNeatDecoder(2, 5, 5, true),
        () => CreateHyperNeatDecoder(2, 5, 3, true),
        () => CreateHyperNeatDecoder(2, 5, 3, false)
      };
    }

    private static ConvGenomeFactory CreateNeatFactory(int convCount, int area)
    {
      return new ConvGenomeFactory(convCount, 1, ConvGenomeFactory.CreateLibraryConv(
        DefaultActivationFunctionLibrary.CreateLibraryNeat(SteepenedSigmoid.__DefaultInstance), area), new NeatGenomeParameters
        {
          FeedforwardOnly = true,
          InitialInterconnectionsProportion = 1,
          AddNodeMutationProbability = 0.01,
          AddConnectionMutationProbability = 0.04,
          DeleteConnectionMutationProbability = 0,
          ConnectionWeightMutationProbability = 0.95 / 2,
          NodeAuxStateMutationProbability = 0.95 / 2
        });
    }

    private static ConvGenomeFactory CreateRbfNeatFactory(int convCount, int area)
    {
      return new ConvGenomeFactory(convCount, 1, ConvGenomeFactory.CreateLibraryConv(
        DefaultActivationFunctionLibrary.CreateLibraryRbf(SteepenedSigmoid.__DefaultInstance, 0, 0.1), area), new NeatGenomeParameters
        {
          FeedforwardOnly = true,
          InitialInterconnectionsProportion = 1,
          AddNodeMutationProbability = 0.01,
          AddConnectionMutationProbability = 0.04,
          DeleteConnectionMutationProbability = 0,
          ConnectionWeightMutationProbability = 0.95 / 2,
          NodeAuxStateMutationProbability = 0.95 / 2
        });
    }

    private static IGenomeDecoder<NeatGenome, IBlackBox> CreateHyperNeatDecoder(int from, int span, int offset, bool withHidden2)
    {
      var input = new SubstrateNodeSet();
      var output = new SubstrateNodeSet();
      var hidden1 = new SubstrateNodeSet();
      var hidden2 = new SubstrateNodeSet();

      double half = (WindowSize - 1) / 2.0;

      uint id = 1;
      for (int y = 0; y < WindowSize; ++y)
      {
        for (int x = 0; x < WindowSize; ++x)
        {
          input.NodeList.Add(new SubstrateNode(id++, new double[] { (x - half) / half, (y - half) / half, -1 }));
        }
      }

      output.NodeList.Add(new SubstrateNode(id++, new double[] { 0, 0, 1 }));

      double zHidden1 = withHidden2 ? -1.0 / 3 : 0;
      double xyHidden1 = span % 2 == 0 ? 0.5 : 0;

      for (int y = from; y < WindowSize; y += offset)
      {
        for (int x = from; x < WindowSize; x += offset)
        {
          hidden1.NodeList.Add(new SubstrateNode(id++,
            new double[] { (x + xyHidden1 - half) / half, (y + xyHidden1 - half) / half, zHidden1 }));
        }
      }

      for (int y = -1; y <= 1; y += 1)
      {
        for (int x = -1; x <= 1; x += 1)
        {
          hidden2.NodeList.Add(new SubstrateNode(id++, new double[] { x, y, 1.0 / 3 }));
        }
      }

      var sets = new List<SubstrateNodeSet> { input, output, hidden1 };
      var mappings = new List<NodeSetMapping>
      {
        NodeSetMapping.Create(0, 2, new AreaMappingFunction((double)span / WindowSize, (double)span / WindowSize)),
      };

      if (withHidden2)
      {
        sets.Add(hidden2);
        mappings.Add(NodeSetMapping.Create(2, 3, (double?)null));
        mappings.Add(NodeSetMapping.Create(3, 1, (double?)null));
      }
      else
      {
        mappings.Add(NodeSetMapping.Create(2, 1, (double?)null));
      }

      var substrate = new Substrate(sets, DefaultActivationFunctionLibrary.CreateLibraryNeat(SteepenedSigmoid.__DefaultInstance), 0, 0.1, 5, mappings);
      return new HyperNeatDecoder(substrate, NetworkActivationScheme.CreateAcyclicScheme(), NetworkActivationScheme.CreateAcyclicScheme(), false);
    }

    private static CppnGenomeFactory CreateHyperNeatFactory()
    {
      return new CppnGenomeFactory(6, 2, DefaultActivationFunctionLibrary.CreateLibraryCppn(), new NeatGenomeParameters
      {
        FeedforwardOnly = true,
        InitialInterconnectionsProportion = 1,
        AddNodeMutationProbability = 0.01,
        AddConnectionMutationProbability = 0.02,
        DeleteConnectionMutationProbability = 0,
        ConnectionWeightMutationProbability = 0.97
      });
    }
  }
}
