﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace FaceDetection.Evolution
{
  public sealed class Clusterer
  {
    private readonly int clustersCount;

    public Clusterer(int clustersCount)
    { 
      this.clustersCount = clustersCount;
    }

    public List<Cluster> InitClusters(List<Genome> genomes, Random random)
    {
      var clusters = new List<Cluster>() { new Cluster(genomes.GetRandom(random)) };
      while (clusters.Count != clustersCount)
      {
        var distances = genomes.Select(g => clusters.Min(c => Vector.Distance(g, c.Center))).ToList();
        var comulative = new double[distances.Count];
        comulative[0] = distances[0];
        for (int i = 1; i < distances.Count; ++i)
        {
          comulative[i] = comulative[i - 1] + distances[i];
        }

        var target = random.NextDouble() * comulative[comulative.Length - 1];
        for (int i = 0; i < distances.Count; ++i)
        {
          if (comulative[i] > target)
          {
            clusters.Add(new Cluster(genomes[i]));
            break;
          }
        }
      }

      UpdateClusters(clusters, genomes);
      return clusters;
    }

    public void UpdateClusters(List<Cluster> clusters, List<Genome> genomes)
    {
      const int maxTries = 5;
      for (int i = 0; i < maxTries; ++i)
      {
        bool changed = false;
        foreach (var genome in genomes)
        {
          var closest = clusters.MinBy(c => Vector.Distance(c.Center, genome));
          changed |= closest != genome.Cluster;
          genome.Cluster = closest;
        }

        if (!changed)
        {
          break;
        }

        foreach (var cluster in clusters)
        {
          cluster.Clear();
        }
        foreach (var genome in genomes)
        {
          genome.Cluster.Add(genome);
        }

        if (clusters.Count != clustersCount)
        {
          var genome = genomes.MinBy(g => -Vector.Distance(g, g.Cluster.Center));
          genome.Cluster.Remove(genome);
          clusters.Add(new Cluster(genome));
        }
      }
    }
  }
}
